"""Datapackage class.

The class contains methods for building and publishing a datapackage.
"""
import datetime
import io
import json
import os
from typing import Any, Dict, List

import pandas as pd
import plotly.io as pio

import deetly.dcat
import deetly.utils

from enum import Enum


class DatapackageError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        print("calling str")
        if self.message:
            return "DatapackageError, {0} ".format(self.message)
        else:
            return "DatapackageError has been raised"


class Datapackage:
    """Datapackage resource.

    Attributes:
        id: The id of the datapackage.
        name: The name of the datapackage.
    """

    def __init__(self, metadata: Dict) -> None:
        """Constructor."""
        self.resources = []
        self.views: List = []
        self.metadata: Dict = self._create_datapackage(dict(metadata))

    def _create_datapackage(self, metadata: Dict, validate: str = "None") -> Dict:
        """Create datapackage from metadata."""
        now = datetime.datetime.today().isoformat()
        metadata["issued"] = metadata.get("created", now)
        metadata["modified"] = now

        if validate == "dcat": 
            ds = deetly.dcat.Dataset(metadata)
            ds.validate()

        return metadata

    @property
    def id(self) -> str:
        """Datapackage id."""
        _id = self.metadata.get("id", deetly.utils.get_id_from_metadata(self.metadata))
        self.metadata["id"] = _id
        return _id

    @property
    def name(self) -> str:
        """Datapackage name."""
        return self.metadata.get("name", None)

    def plot(
        self,
        spec_type: str,
        spec: Dict,
        name: str,
        description: str,
        resources: str or List = [],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Adds a plot/figure/view to the package.

        Creates a dict containg layout, data and metadata

        Args:
            spec_type: For example 'ploty', 'vega' or other type.
            spec: The JSON spec for the view.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
            args: optional additional argument list.
            kwargs: optional keyworded, argument list.
        """

        view = {
            "name": name,
            "description": description,
            "specType": spec_type,
            "spec": spec,
            "resources": resources,
        }

        for key, value in kwargs.items():
            view[key] = value

        self.views.append(view)

    def add(self, type: str, content: Any, name: str, description: str = "") -> None:
        """Adds a view view to the package.

        Args:
            content: The view content.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
            type: The type of view
        """ 

        if type == 'plotly':
            return self.plotly(content, name, description)

        if type == 'altair':
            return self.altair(content, name, description)
        
        if type == 'link':
            return self.link(content, name, description)

        if type == 'text':
            return self.markdown(content, name, description)

        if type == 'deck':
            return self.deck(content, name, description)

        if type == 'pydeck':
            return self.pydeck(content, name, description)

        if type == 'barlist':
            return self.barlist(content, name, description)

        if type == 'markdown':
            return self.markdown(content, name, description)
        
        if type == 'table':
            return self.table(content, name, description)

        if type == 'vega':
            return self.vega(content, name, description)

        if type == 'pdf':
            return self.pdf(content, name, description)

        if type == 'forcegraph':
            return self.forcegraph(content, name, description)

        if type == 'cytoscape':
            return self.cytoscape(content, name, description)

        if type == 'upload':
            return self.upload(content, name, description)

        if type == 'echart':
            return self.echart(content, name, description)

        print("No viewer available for " + name)

    def echart(self, fig: object, name: str, description: str, layout: Dict = {}) -> None:
        """Adds a EChart view to the package.

        Creates a dict containg layout, data and metadata

        Args:
            fig: The echart figure.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """
        
        option = fig.dump_options_with_quotes()
        option = option.replace('\n', '') 
        spec = {"option": json.loads(option)}
    
        self.plot("echart", spec, name, description)


    def plotly(self, fig: Dict, name: str, description: str) -> None:
        """Adds a plotly view to the package.

        Creates a dict containg layout, data and metadata

        Args:
            fig: The Plotly figure.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        self.plot("plotly", pio.to_json(fig), name, description)

    def altair(
        self, fig: Any, name: str, description: str, *args: Any, **kwargs: Any
    ) -> None:
        """Adds a Altair view to the package.

        Creates a dict containg layout, data and metadata

        Args:
            fig: The Altair figure.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
            args: optional additional argument list.
            kwargs: optional keyworded, argument list.
        """

        self.plot("vega", fig.to_dict(), name, description, *args, **kwargs)

    def vega(
        self, fig: Any, name: str, description: str, *args: Any, **kwargs: Any
    ) -> None:
        """Adds a Vega view to the package.

        Creates a dict containg layout, data and metadata

        Args:
            fig: The Vega chart.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
            args: optional additional argument list.
            kwargs: optional keyworded, argument list.
        """

        self.plot("vega", fig.to_dict(), name, description, *args, **kwargs)

    def deck(
        self, fig: Any, name: str, description: str, *args: Any, **kwargs: Any
    ) -> None:
        """Adds a Deck view to the package.

        Creates a deck.gl view

        Args:
            fig: The Deck chart.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
            args: optional additional argument list.
            kwargs: optional keyworded, argument list.
        """

        self.plot("deck", fig, name, description, *args, **kwargs)

    def pydeck(
        self, fig: Any, name: str, description: str, *args: Any, **kwargs: Any
    ) -> None:
        """Adds a PyDeck view to the package.

        Creates a dpydeck view

        Args:
            fig: The pydeck chart.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
            args: optional additional argument list.
            kwargs: optional keyworded, argument list.
        """

        self.plot("deck", json.loads(fig.to_json()), name, description, *args, **kwargs)

    def barlist(self, fig: Any, name: str, description: str = "") -> None:
        """Adds a Barlist view to the package.

        Creates barlist view

        Args:
            fig: The specification.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        self.plot("barlist", fig, name, description)

    def forcegraph(self, fig: Any, name: str, description: str = "") -> None:
        """Adds a ForceGraph view to the package.

        Creates forcegraph view

        Args:
            fig: The specification.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        self.plot("forcegraph", fig, name, description)

    def cytoscape(self, fig: Any, name: str, description: str = "") -> None:
        """Adds a Cytoscape graph view to the package.

        Creates cytoscape view

        Args:
            fig: The specification.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        self.plot("cytoscape", fig, name, description)

    def graphviz(self, fig: Any, name: str, description: str = "", *args: Any,
        **kwargs: Any) -> None:
        """Adds a Graphviz graph view to the package.

        Creates graphviz view

        Args:
            fig: The specification.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        fig = {"data": fig}

        self.plot("graphviz", fig, name, description)

    def pdf(
        self, resource: str or object, name: str = "", description: str = ""
    ) -> None:
        """Adds a PDF view to the package.

        Creates pdf view

        Args:
            resource: Name of the resource.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        self.plot("pdf", {}, name if name else resource, description, resource)

    def table(self, resource: str, name: str = "", description: str = "") -> None:
        """Adds a table view to the package.

        Creates table view

        Args:
            resource: Name of the resource.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        self.plot("table", {}, name if name else resource, description, resource)

    def markdown(self, markdown: str, name: str, description: str = "") -> None:
        """Adds a Markdown view to the package.

        Creates markdown view

        Args:
            fig: The specification.
            name: View name.
            description: Description of the view. Can be plain text or Markdown.
        """

        spec = {"markdown": markdown}

        self.plot("markdown", spec, name, description)

    def df(self, df: pd.DataFrame, name: str, description: str = "") -> None:
        """Add a pandas dataframe to the package.

        Saves pandas dataframe to gzip'ed csv file.

        Args:
            df: The pandas dataframe
            name: The name of the datasett
            description: Description of the data. Can be plain text or Markdown

        Raises:
            TypeError: The first parameter (df) must be of type pandas.Dataframe().
        """

        if not isinstance(df, pd.DataFrame):
            raise TypeError(
                f"The first parameter df must be of type pandas.Dataframe()"
            )

        fields = []
        for name, dtype in zip(df.columns, df.dtypes):

            if str(dtype) == "object":
                dtype = "string"
            if str(dtype) == "int64":
                dtype = "number"
            if str(dtype) == "float64":
                dtype = "number"
            if str(dtype) == "bool":
                dtype = "boolean"
            if str(dtype) == "datetime64":
                dtype = "string"
            if str(dtype) == "timedelta[ns]":
                dtype = "string"
            if str(dtype) == "category":
                dtype = "string"
            else:
                dtype = "string"
            fields.append({"name": name, "type": dtype})

        output = io.StringIO()
        df.to_csv(output, index=False)
        data = output.getvalue()
        output.close()

        name = name.replace(" ", "_")

        resource = {
            "name": name,
            "description": description,
            "path": f"/resources/{name}.csv.gz",
            "format": "csv",
            "dsv_separator": ",",
            "mediatype": "text/csv",
            "schema": {"fields": fields},
            "data": data,
        }
        self.resources.append(resource)

    def url(
        self,
        url: str,
        name: str,
        description: str = "",
        format: str = "pdf",
        mediatype: str = "application/pdf",
        resourcetype: str = "url",
    ) -> None:
        """Add a url to the package.

        Args:
            url: Full path
            name: The name of file
            description: Description of the data. Can be plain text or Markdown

        Raises:
            TypeError: The first parameter (url) must be of type str.
        """

        if not isinstance(url, str):
            raise TypeError(f"The first parameter url must be of type str")

        name = name.replace(" ", "_")

        resource = {
            "name": name,
            "description": description,
            "path": url,
            "format": format,
            "mediatype": mediatype,
            "type": resource,
        }
        self.resources.append(resource)

    def link(self, url: str, name: str, description: str = "") -> None:
        """Add a link to the package as a resource

        Args:
            url: The URL of the link
            name: Name 
            description: Description of the data. Can be plain text or Markdown
        """

        (
            resource_name,
            resource_extension,
        ) = deetly.utils.get_name_and_extension_from_url(url)

        link = {
            "name": name,
            "description": description if description else name,
            "path": url,
            "format": resource_extension.replace(".",""),
            "mediatype": deetly.utils.get_media_type(resource_extension),
            "type": "link",
        }

        self.resources.append(link)

    def upload(self, url: str, name: str = "", description: str = "") -> None:
        """Upload a file and add the file as downloadable dataresource to the package.

        Args:
            url: The URL of the file
            name: Name 
            description: Description of the file. Can be plain text or Markdown
        """

        (
            resource_name,
            resource_extension,
        ) = deetly.utils.get_name_and_extension_from_url(url)

        name = name.replace(" ", "_")

        resource = {
            "name": name,
            "description": description if description else name,
            "url": url,
            "format": resource_extension,
            "mediatype": deetly.utils.get_media_type(resource_extension),
            "data": None,
        }

        self.resources.append(resource)

    def _upload_resources(self, space: str, token: str) -> List:
        resources = []

        # upload resources
        for resource in self.resources:
            if resource.get("data", None) is not None:
                public_url = deetly.utils.upload_file(
                    space,
                    token,
                    self.id,
                    f'resources/{resource["name"]}',
                    resource["mediatype"],
                    resource["data"],
                )
            else:
                public_url = deetly.utils.upload_from_url(
                    space,
                    token,
                    self.id,
                    f'resources/{resource["name"]}',
                    resource["mediatype"],
                    resource["path"],
                )

            _resource = {}
            for key, value in resource.items():
                if key not in ["data"]:
                    _resource[key] = value

            _resource["description"] = resource.get("description", "")
            _resource["url"] = json.loads(public_url).get(
                "path", resource.get("path", "")
            )
            _resource["type"] = resource.get("type", "dataset")
            resources.append(_resource)

        return resources

    def _get_content(self, space: str, token: str) -> List:
        content = []
        for view in self.views:
            for key, value in view.items():
                if key == "name":
                    content.append(
                        {
                            "type": "view",
                            "name": value,
                            "description": view.get("description", ""),
                        }
                    )

        return content

    def publish(self, space: str = None, token: str = None) -> None:
        """Publishes the package to storage and the metadata to elastic seach.

        Args:
            space: The id of the space to publish to.
                Defaults to environment variable [DEETLY_SPACE]
            token: The users token.
                Defaults to environment variable [DEETLY_TOKEN]
        """

        if space is None:
            space = os.environ["DEETLY_SPACE"]

        if space is None:
            msg = """The space id must be provided either as parameter
                    or set with environment variable: DEETLY_SPACE
            """
            print(msg)
            raise DatapackageError(msg)

        if token is None:
            token = os.environ["DEETLY_TOKEN"]

        if token is None:
            msg = """Your token must be provided either as parameter
                    or set with environment variable: DEETLY_TOKEN
            """
            print(msg)
            raise DatapackageError(msg)

        self.metadata["id"] = self.id
        self.metadata["views"] = self.views

        self.metadata["resources"] = self._upload_resources(space, token)
        self.metadata["content"] = self._get_content(space, token)

        datapackage = json.dumps(self.metadata)

        # upload datapackage
        response = deetly.utils.upload_file(
            space, token, self.id, "datapackage.json", "application/json", datapackage,
        )
        datapackage_url = json.loads(response).get("url", "")
        print(f"View: https://public.deetly.com/{space}/{self.id}", "\n")

        print(f"Metadata: {datapackage_url}", "\n")

        if len(self.metadata["resources"]) > 0:
            print(f"Resources:", "\n")
            for resource in self.metadata["resources"]:
                print(resource["url"], "\n")

        # extract metadata for elastic search
        metadata = {}
        for key, value in self.metadata.items():
            if key not in ["views"]:
                metadata[key] = value

        metadata["url"] = datapackage_url
        metadata["space"] = space

        res = json.loads(deetly.utils.index_document(token, space, metadata))
        body = res.get("body", None)

        if not body:
            print(f"Error updating metadata index: {res}", "\n")

        result = body.get("result", None)

        if not result:
            print("\n", f"Error updating metadata index: {res}")



from schema import Schema, And, Or, Use, Optional
from typing import Any, Dict, List
from datetime import datetime
from dateutil import parser

is_datetime_string = lambda date: isinstance(parser.parse(date), datetime)

def validate_datapackage_minimal(package: Dict):

    package_schema = Schema({
        'name': And(str, len),
        'description': And(str, len),
        'author': And(str, len),
        'keywords': And([str]),
        'theme': And(str, len),
        'license': And(str, len)
        })

    return   Schema(package_schema).validate(package)

def validate_datapackage_dcat2(package: Dict):

    package_schema = Schema({
        'type': And(str, len),
        'title': And(str, len),
        'description': And(str, len),
        'issued': And(Use(is_datetime_string)),
        "modified": And(Use(is_datetime_string)),
        "language": Or(str,[str]),
        "keyword": Or(str,[str]),
        "publisher": And(str, len),
        "distribution": [object]
        })

    return   Schema(package_schema).validate(package)


"""Deetly utils.

Utility functions
"""
import hashlib
import re
from typing import Dict
import uuid

import pandas
import pyarrow
import requests
import urllib3
import mimetypes

api_url = "https://us-central1-mesiqi-238408.cloudfunctions.net/deetlyAPI"
# api_url = "http://localhost:8080"


def get_schema(
    df: pandas.DataFrame,
    resource_name: str,
    resource_description: str,
    format: str,
    compress: str,
    dsv_separator: str,
    spec: dict,
) -> dict:
    """Get Schema from Pandas dataframe.

    Args:
        df: pandas dataframe.
        resource_name: resoure name.
        resource_description: resource description.
        format: file format.
        compress: boolean. True for compressed files.
        dsv_separator: dsv/csv separator.
        spec: figure spec.

    Returns:
        dict
    """

    fields = []

    for name, dtype in zip(df.columns, df.dtypes):
        if str(dtype) == "object":
            dtype = "string"
        else:
            dtype = "number"

        description = ""
        if spec and spec.get("fields"):
            spec_fields = spec["fields"]
            for field in spec_fields:
                if field["name"] == name:
                    description = field["description"]

        fields.append({"name": name, "description": description, "type": dtype})

    mediatype = get_media_type(format)

    return {
        "name": resource_name,
        "description": resource_description,
        "format": format,
        "dsv_separator": dsv_separator,
        "compressed": compress,
        "mediatype": mediatype,
        "schema": {"fields": fields},
        "spec": spec,
    }


def get_media_type(fmt: str) -> str:
    """Guess media type from file type.

    Args:
        fmt: file type

    Returns:
        String
    """

    return mimetypes.types_map[fmt]


def get_resource_path(
    path: str, resource_name: str, fmt: str, compress: bool = False
) -> str:
    """Get path of resource file.

    Args:
        path: root path
        resource_name: name of the resource
        fmt: file type
        compress: boolean. True for compressed files

    Returns:
        String
    """
    if compress:
        return f"{path}/resources/{resource_name}.{fmt}.gz"
    else:
        return f"{path}/resources/{resource_name}.{fmt}"


def verify_add_resource_input_types(
    df: pandas.DataFrame, dataset_name: str, dataset_description: str
) -> None:
    if not isinstance(df, pandas.DataFrame):
        raise TypeError(f"df must be of type pandas.Dataframe()")
    if not isinstance(dataset_name, str):
        raise TypeError(f"dataset_name must be of type string")
    if not isinstance(dataset_description, str):
        raise TypeError(f"dataset_description must be of type string")


def get_id_from_metadata(metadata: Dict) -> str:
    """Get id from metadata dict if availble.

    If id not available then a new id is inserted.
    The id is calculated as a hash to metadata dict.

    Args:
        metadata: metadata dict

    Returns:
        String
    """

    author = metadata.get("author", None)
    name = metadata.get("name", None)
    title = metadata.get("title", None)
    theme = metadata.get("theme", None)
    # created = metadata.get("created", datetime.datetime.today().isoformat())

    id_string = "-".join(filter(None, (author, name, title, theme)))
    if id_string:
        # pylint: disable=S303
        hash_object = hashlib.md5(id_string.encode())
        dp_id = hash_object.hexdigest()
        return re.sub("[^0-9a-z]+", "-", dp_id.lower())
    else:
        return str(uuid.uuid4())


def get_name_and_extension_from_url(resource_url: str) -> [str, str]:
    """Extract name and extension from url.

    Args:
        resource_url: resource URL

    Returns:
        String

    Raises:
        ValueError: If the input string is not a valid url format.
    """

    parsed_url = urllib3.util.parse_url(resource_url)

    if not parsed_url.scheme == "https" and not parsed_url.scheme == "http":
        raise ValueError(
            f"Remote resource needs to be a web address, scheme is {parsed_url.scheme}"
        )

    resource = parsed_url.path.split("/")[-1]

    name = resource.split(".", 1)[0]

    if len(name) == 0:
        raise ValueError(f"Url does not contain a filename")

    mimetype, encoding = mimetypes.guess_type(resource_url)
    extension = mimetypes.guess_extension(mimetype)

    if extension == None:
         raise ValueError(f"Unable to guess extension from mimetype")

    if len(extension) == 0:
        raise ValueError(f"Url does not contain an extension")

    return name, extension


def serialize_table(df: pandas.DataFrame, compress: bool = False) -> bytearray:
    """Serialize pandas dataframe to arrow buffer.

    Args:
        df: pandas dataframe.
        compress: boolean. True for compressed files.

    Returns:
        Bytearray
    """
    ser = pyarrow.serialize(df).to_buffer()
    return ser


def read_file(url: str) -> str:
    """Download file and return contents as a string.

    Args:
        url: File url

    Returns:
        str
    """

    try:
        c = requests.get(url)
        c.raise_for_status()
    except requests.exceptions.RequestException as e:
        print(e)
        return ""

    return c.content


def upload_file(
    space: str,
    token: str,
    package_id: str,
    file_name: str,
    content_type: str,
    content: str,
) -> str:
    """Upload file to storage.

    Args:
        space: Space Id
        token: Access token
        package_id: Datapackage id
        file_name: Name of the file
        content_type: Content type
        content: File content as string

    Returns:
        str
    """

    res = requests.post(
        f"{api_url}/store",
        json={
            "package_id": package_id,
            "content": content,
            "name": file_name,
            "content_type": content_type,
            "space": space,
            "token": token,
        },
    )

    return res.text


def upload_from_url(
    space: str,
    token: str,
    package_id: str,
    file_name: str,
    content_type: str,
    url: str,
) -> str:
    """Upload file to storage.

    Args:
        space: Space Id
        token: Access token
        package_id: Datapackage id
        file_name: Name of the file
        content_type: Content type
        url: Source url

    Returns:
        str
    """

    res = requests.post(
        f"{api_url}/upload",
        json={
            "package_id": package_id,
            "url": url,
            "name": file_name,
            "content_type": content_type,
            "space": space,
            "token": token,
        },
    )

    return res.text


def index_document(space: str, token: str, doc: Dict,) -> str:
    """Add/update Elastic search item.

    Args:
        token: Access token
        space: Space id
        doc: Document (JSON) to add to index

    Returns:
        str
    """

    res = requests.post(
        f"{api_url}/index", json={"space": space, "token": token, "doc": doc}
    )

    return res.text



"""Deetly dcat.

Utility functions for validating dcat formated metadata
"""
from collections import defaultdict
from typing import Any, Dict, List

import isodate


class ValidationError:
    """ValidationError base class."""

    ERROR_OTHER = 100
    ERROR_INVALID_VALUE = 101
    ERROR_MISSING_PROPERTY = 102
    ERROR_VALIDATION_STRING_TYPE = 103
    ERROR_VALIDATION = 104
    ERROR_INVALID_PROPERTY = 105
    WARNING_MISSING_RECOMMENDED_PROPERTY = 106

    def to_string(self, code: str, fields: List) -> str:
        """Errors to string."""

        items = []
        for field in fields:
            if field.get("validator"):
                items.append(f"{field['validator'].hint}:{field['field']}")
            else:
                items.append(f"{field['field']}")
        if code == ValidationError.ERROR_MISSING_PROPERTY:
            return f"Missing mandatory props(s): {items}"
        if code == ValidationError.WARNING_MISSING_RECOMMENDED_PROPERTY:
            return f"Missing recommended props(s): {items}"
        elif code == ValidationError.ERROR_INVALID_VALUE:
            return f"Not valid: {field}"
        elif code == ValidationError.ERROR_OTHER:
            return f"Could not be validated: {items}"
        elif code == ValidationError.ERROR_VALIDATION:
            return f"Invalid type: {items}"
        elif code == ValidationError.ERROR_INVALID_PROPERTY:
            return f"Property not in DCAT schema: {items}"


class Validator(object):
    """Validator base class."""

    hint = "Not implemented validator"

    def validate(self) -> None:
        """Validate method in base class."""
        raise NotImplementedError()


class StringValidator(Validator):
    """Validator for string properties."""

    hint = "String"

    def validate(self, value: str) -> None:
        """Validate input."""
        if not isinstance(value, str):
            raise ValidationError.ERROR_VALIDATION


class IntValidator(Validator):
    """Validator for integer properties."""

    hint = "Integer"

    def validate(self, value: int) -> None:
        """Validate input."""
        if not isinstance(value, int):
            raise ValidationError.ERROR_VALIDATION


class DateValidator(Validator):
    """Validator for date or datetime properties."""

    hint = "ISO date or datetime"

    def validate(self, value: str) -> None:
        """Validate input."""
        try:
            isodate.parse_date(value)
        except Exception:
            try:
                isodate.parse_datetime(value)
            except Exception:
                raise ValidationError.ERROR_VALIDATION


class StringOrListOfStringsValidator(Validator):
    """Validator for string or list of strings properties."""

    hint = "String or list of strings"

    def validate(self, value: Any) -> None:
        """Validate input."""
        if isinstance(value, list):
            for val in value:
                if not isinstance(val, str):
                    raise ValidationError.ERROR_VALIDATION
        else:
            if not isinstance(value, str):
                raise ValidationError.ERROR_VALIDATION


class ContactValidator(Validator):
    """Validator for contact dict properties."""

    hint = "Dict"

    def validate(self, value: dict) -> None:
        """Validate input."""
        if not isinstance(value, dict):
            raise ValidationError.ERROR_VALIDATION

        if not isinstance(value["name"], str):
            raise ValidationError.ERROR_VALIDATION

        if not isinstance(value["mbox"], str):
            raise ValidationError.ERROR_VALIDATION


class DistributionValidator(Validator):
    """Validator for Distribution properties."""

    hint = "Distribution validator"

    def validate(self, value: list) -> None:
        """Validate input."""
        if not isinstance(value, list):
            raise ValidationError.ERROR_VALIDATION

        for dist in value:
            dist = Distribution(dist)
            dist.validate()


class DcatModel(object):
    """DCAT model."""

    def __init__(self, metadata: dict) -> None:
        """Constructor."""
        self.metadata = metadata
        self.mandatory = []
        self.recommended = []
        self.optional = []
        self.validators = {
            # 'accessRights':  StringValidator,
            # 'accessService':  StringValidator,
            # 'accessURL':  StringValidator,
            # 'accrualPeriodicity':  StringValidator,
            # 'availability':  StringValidator,
            "byteSize": IntValidator,
            # 'checksum':  StringValidator,
            # 'compressFormat':  StringValidator,
            # 'conformsTo':  StringValidator,
            "contactPoint": ContactValidator,
            # 'creator':  StringValidator,
            "distribution": DistributionValidator,
            # 'downloadURL':  StringValidator,
            "format": StringOrListOfStringsValidator,
            # 'hasPolicy':  StringValidator,
            # 'hasVersion':  StringValidator,
            # 'identifier':  StringValidator,
            "isReferencedBy": StringValidator,
            # 'isVersionOf':  StringValidator,
            "issued": DateValidator,
            "keyword": StringOrListOfStringsValidator,
            # 'landingPage':  StringValidator,
            "language": StringOrListOfStringsValidator,
            # 'license':  StringValidator,
            # 'mediaType':  StringValidator,
            "modified": DateValidator,
            # 'packageFormat':  StringValidator,
            "page": StringValidator,
            # 'provenance':  StringValidator,
            "publisher": ContactValidator,
            # 'qualifiedAttributionwasGeneratedBy':  StringValidator,
            # 'qualifiedRelation':  StringValidator,
            # 'relation':  StringValidator,
            # 'rights':  StringValidator,
            # 'sample':  StringValidator,
            # 'source':  StringValidator,
            # 'spatial':  StringValidator,
            "spatialResolutionInMeters": IntValidator,
            # 'status':  StringValidator,
            # 'temporal':  StringValidator,
            # 'temporalResolution':  StringValidator,
            "theme": StringOrListOfStringsValidator,
            # 'type':  StringValidator,
            # 'versionInfo':  StringValidator,
            # 'versionNotes':  StringValidator,
        }

    def get_validator(self, key: Any) -> Validator:
        """Get validator for key."""
        return self.validators.get(key, StringValidator)

    def get_properties(self) -> List:
        """Get list of all properties."""
        return [*self.mandatory, *self.recommended, *self.optional]

    def validate_content(self, errors: Dict) -> None:
        """Check if mandatory and recommended properties are included."""
        for field in self.mandatory:
            if field not in self.metadata.keys():
                errors[ValidationError.ERROR_MISSING_PROPERTY].append({"field": field})

        for field in self.recommended:
            if field not in self.metadata.keys():
                errors[ValidationError.WARNING_MISSING_RECOMMENDED_PROPERTY].append(
                    {"field": field}
                )

    def validate(self) -> List:
        """Validate."""
        errors = defaultdict(list)

        self.validate_content(errors)

        for field, value in self.metadata.items():
            if field not in self.get_properties():
                errors[ValidationError.ERROR_INVALID_PROPERTY].append({"field": field})

            try:
                validator = self.get_validator(field)
            except KeyError:
                errors[ValidationError.ERROR_INVALID_VALUE].append(
                    {"field": field, "validator": validator}
                )
            except Exception:
                errors[ValidationError.ERROR_OTHER].append(
                    {"field": field, "validator": validator}
                )

            try:
                validator.validate(value)
            except Exception:
                errors[ValidationError.ERROR_VALIDATION].append(
                    {"field": field, "validator": validator}
                )

        for key, value in errors.items():
            print(ValidationError().to_string(key, value))

        return errors


class Dataset(DcatModel):
    """DCAT Dataset."""

    def __init__(self, metadata: Dict) -> None:
        """Constructor."""
        super(Dataset, self).__init__(metadata)
        self.mandatory = ["description", "title"]

        self.recommended = [
            "contactPoint",
            "distribution",
            "keyword",
            "publisher",
            "spatial",
            "temporal",
            "theme",
        ]

        self.optional = [
            "identifier",
            "sample",
            "versionNotes",
            "landingPage",
            "spatialResolutionInMeters",
            "temporalResolution",
            "qualifiedRelation",
            "accessRights",
            "accrualPeriodicity",
            "conformsTo",
            "creator",
            "hasVersion",
            "isReferencedBy",
            "isVersionOf",
            "identifier",
            "issued",
            "language",
            "modified",
            "provenance",
            "relation",
            "source",
            "type",
            "page",
            "versionInfo",
            "qualifiedAttribution" "wasGeneratedBy",
        ]


class Distribution(DcatModel):
    """DCAT Distibution."""

    def __init__(self, metadata: Dict) -> None:
        """Constructor."""
        super(Distribution, self).__init__(metadata)

        self.mandatory = ["accessURL"]

        self.recommended = ["availability", "description", "format", "license"]

        self.optional = [
            "status",
            "accessService",
            "byteSize",
            "compressFormat",
            "downloadURL",
            "mediaType",
            "packageFormat",
            "spatialResolutionInMeters",
            "temporalResolution",
            "conformsTo",
            "issued",
            "language",
            "modified",
            "rights",
            "title",
            "page",
            "hasPolicy",
            "checksum",
        ]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['deetly']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=3.4.0,<4.0.0',
 'isodate>=0.6.0,<0.7.0',
 'pandas>=1.0.1,<2.0.0',
 'plotly>=4.6.0,<5.0.0',
 'pyarrow>=0.16.0,<0.17.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests>=2.22.0,<3.0.0',
 'schema>=0.7.2,<0.8.0']

entry_points = \
{'console_scripts': ['deetly = deetly.console:main']}

setup_kwargs = {
    'name': 'deetly',
    'version': '0.0.9',
    'description': 'toolkit for creating data packages',
    'long_description': None,
    'author': 'Paul Bencze',
    'author_email': 'paul@idelab.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
